package org.xbib.interlibrary.z;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.security.SecureRandom;
import java.text.Normalizer;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

import org.xbib.datastructures.common.MultiMap;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.datastructures.tiny.TinyList;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.datastructures.tiny.TinyMultiMap;
import org.xbib.interlibrary.action.avail.AbstractAvailService;
import org.xbib.interlibrary.action.avail.AvailRequest;
import org.xbib.interlibrary.action.avail.AvailResponseBuilder;
import org.xbib.interlibrary.action.avail.AvailServiceResponse;
import org.xbib.interlibrary.action.avail.AvailServiceResponseBuilder;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.library.DefaultLibraryService;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.api.library.Mode;
import org.xbib.interlibrary.common.util.InputStreamHelper;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.z3950.client.api.Client;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;

public abstract class AbstractZAvailService extends AbstractAvailService {

    private static final Random RANDOM = new SecureRandom();

    private static final Logger logger = Logger.getLogger(AbstractZAvailService.class.getName());

    protected final ZClientHelper zClientHelper;

    protected final MarcHelper marcHelper;

    protected final ReentrantLock lock;

    protected AbstractZAvailService(ServiceArguments arguments) {
        super(arguments);
        this.zClientHelper = new ZClientHelper(arguments);
        this.marcHelper = MarcHelper.INSTANCE;
        this.lock = new ReentrantLock();
    }

    @Override
    public void execute(AvailRequest availRequest, AvailResponseBuilder availResponse) {
        AvailServiceResponseBuilder availServiceResponse = AvailServiceResponse.builder();
        if (!availRequest.isValidated()) {
            logger.log(Level.FINE, "request not validated, returning empty response");
            availServiceResponse.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
            return;
        }
        availServiceResponse.setRequest(availRequest);
        availServiceResponse.setService(getName());
        try {
            availServiceResponse.setRequest(availRequest);
            MultiMap<String, LibraryService> multiMap = new TinyMultiMap<>();
            fetchAvailServices(availRequest, availServiceResponse, multiMap);
            toResult(availRequest, availServiceResponse, multiMap);
        } catch (Throwable e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            availServiceResponse.setStatus(getName(), Response.Status.ERROR);
            availServiceResponse.setThrowable(getName(), e.getMessage(), e);
            throw new IllegalStateException(e);
        } finally {
            availResponse.add(availServiceResponse);
        }
    }

    @Override
    public void close() throws IOException {
        zClientHelper.close();
    }

    @Override
    public String getURL(String id) {
        return arguments.getServiceSettings("url", "");
    }

    @Override
    public String getLabel() {
        return arguments.getServiceSettings("label", getName().toUpperCase(Locale.ROOT));
    }

    @Override
    public String getSortKey() {
        return arguments.getServiceSettings("sort", "00");
    }

    protected String getResolveCQL(AvailRequest availRequest) {
        String source = availRequest.getBibliographicDescription().getSource();
        if (source == null) {
            return null;
        }
        source = supports(source);
        if (source == null) {
            return null;
        }
        String sourceId = availRequest.getBibliographicDescription().getSourceId();
        if (sourceId == null) {
            return null;
        }
        return sourceIdCQL(normalizeSourceId(sourceId));
    }

    protected String getResolvePQF(AvailRequest availRequest) {
        String source = availRequest.getBibliographicDescription().getSource();
        if (source == null) {
            return null;
        }
        source = supports(source);
        if (source == null) {
            return null;
        }
        String sourceId = availRequest.getBibliographicDescription().getSourceId();
        if (sourceId == null) {
            return null;
        }
        return sourceIdPQF(normalizeSourceId(sourceId));
    }

    protected abstract String supports(String source);

    protected String normalizeSourceId(String sourceId) {
        return sourceId.toLowerCase(Locale.ROOT).trim().replace("-", "");
    }

    protected String sourceIdCQL(String sourceId) {
        return "bib.identifierLocalNumber = " + sourceId;
    }

    protected String sourceIdPQF(String sourceId) {
        return getClientAttributes("bib.identifierLocalNumber") + " " + encodeQueryString("bib.identifierLocalNumber", sourceId);
    }

    protected void fetchAvailServices(AvailRequest availRequest,
                                      AvailServiceResponseBuilder availResponse,
                                      MultiMap<String, LibraryService> multiMap) {
        try {
            lock.lock();
            String query = getResolveCQL(availRequest);
            if (query == null) {
                return;
            }
            logger.log(Level.FINE, () -> "executing avail request = " + availRequest + " cql = '" + query + "'");
            zClientHelper.open();
            final Client client = zClientHelper.getClient();
            client.searchCQL(query, 0, 1, null,
                (status, total, returned, elapsedMillis) ->
                    logger.log(Level.FINE, "status = " + status + " total = " + total + " returned = " + returned + " millis = " + elapsedMillis),
                record -> {
                    String string = record.toString(StandardCharsets.UTF_8);
                    try (InputStream inputStream = new ByteArrayInputStream(string.getBytes(StandardCharsets.UTF_8))) {
                        String format = zClientHelper.getFormat();
                        String type = zClientHelper.getType();
                        Charset charset = Charset.forName(zClientHelper.getEncoding());
                        Marc.Builder builder = Marc.builder()
                                .setFormat(format)
                                .setType(type)
                                .setInputStream(inputStream)
                                .setCharset(charset);
                        // XML?
                        if ("1.2.840.10003.5.109.10".equals(client.getPreferredRecordSyntax())) {
                            for (MarcRecord marcRecord : builder.xmlIterable()) {
                                avail(availRequest, availResponse, multiMap, marcRecord);
                            }
                        } else {
                            for (MarcRecord marcRecord : builder.iterable()) {
                                avail(availRequest, availResponse, multiMap, marcRecord);
                            }
                        }
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                        availResponse.setStatus(getName(), Response.Status.ERROR);
                        availResponse.setThrowable(getName(), e.getMessage(), e);
                        throw new IllegalStateException(e);
                    }
                },
                () -> availResponse.setStatus(getName(), Response.Status.TIMEOUT));
            zClientHelper.close();
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            availResponse.setStatus(getName(), Response.Status.ERROR);
            availResponse.setThrowable(getName(), e.getMessage(), e);
            throw new IllegalStateException(e);
        } finally {
            lock.unlock();
        }
    }

    protected String getClientAttributes(String key) {
        return zClientHelper.getAttributes().get(key);
    }

    protected boolean isStartOf(String prefix, String string) {
        return string != null && string.startsWith(prefix);
    }

    protected TinyMap.Builder<String, Object> nextElement(TinyList.Builder<TinyMap<String, Object>> list,
                                                          TinyMap.Builder<String, Object> m,
                                                          String key, String value) {
        if (m.containsKey(key)) {
            list.add(m.build());
            m = TinyMap.builder();
        }
        m.put(key, value);
        return m;
    }

    protected String encodeQueryString(String key, String query) {
        String result = query;
        result = result
                .replaceAll("ä", "ae")
                .replaceAll("ö", "oe")
                .replaceAll("ü", "ue")
                .replaceAll("Ä", "Ae")
                .replaceAll("Ö", "Oe")
                .replaceAll("Ü", "Ue")
                .replaceAll("ß", "ss");
        if ("author".equals(key)) {
            // force blank after comma
            result = result.replaceAll("(\\S+),(\\S+)", "\\1, \\2");
        }
        if (key.contains("title")) {
            result = result.replaceAll("(\\S+)-(\\S+)", "\\1#\\2");
        }
        result = Normalizer.normalize(result, Normalizer.Form.NFKD);
        result = result.replaceAll("[^\\x00-\\x7F]", "");
        if ("completetitle".equals(key)) {
            result = "'" + result + "'";
        }
        return result.startsWith("\"") && result.endsWith("\"") ? result : "\"" + result + "\"";
    }

    protected static Integer random() {
        return RANDOM.nextInt();
    }

    protected static Map<String, Object> getMap(String path) {
        try {
            return Json.toMap(InputStreamHelper.readString(AbstractZAvailService.class.getResourceAsStream(path), StandardCharsets.UTF_8));
        } catch (IOException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
            return null;
        }
    }

    protected static Map<String, Mode> getModeMap(String path) {
        try {
            return Json.toMap(InputStreamHelper.readString(AbstractZAvailService.class.getResourceAsStream(path), StandardCharsets.UTF_8))
                    .entrySet().stream()
                    .map(e -> Map.entry(e.getKey(), Mode.fromValue(e.getValue().toString())))
                    .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, (x, y) -> y, HashMap::new));
        } catch (IOException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
            return null;
        }
    }

    /**
     * Extract german dialect of library service.
     *
     * @param marcRecord the MARC record
     * @return settings for service
     */
    protected List<LibraryService> extractGermanInterlibraryServices(AvailRequest availRequest,
                                                                     MarcRecord marcRecord) {
        final List<LibraryService> list = new LinkedList<>();
        String identifier = identifier(marcRecord);
        List<Material> materials = marcHelper.extractMaterial(marcRecord);
        marcRecord.getAll("924").forEach(marcField -> {
            DefaultLibraryService.Builder serviceBuilder = DefaultLibraryService.builder();
            DefaultLibraryService.DefaultLocation location = new DefaultLibraryService.DefaultLocation();
            // ignore indicator 0 or 1
            StringBuilder url = new StringBuilder();
            StringBuilder label = new StringBuilder();
            marcField.getSubfields().forEach(subfield -> {
                switch (subfield.getId()) {
                    case "a":
                        // local ID
                        serviceBuilder.setId(subfield.getValue());
                        break;
                    case "b":
                        // ISIL
                        String isil = subfield.getValue();
                        Library library = configuration.getLibrary(isil);
                        if (library != null) {
                            serviceBuilder.setISIL(library.getISIL());
                            serviceBuilder.setName(library.getName());
                            serviceBuilder.setDomainPriority(domainPriority(availRequest, library));
                            serviceBuilder.setLibraryPriority(libraryPriority(availRequest, library));
                            serviceBuilder.setServicePriority(materials.contains(ONLINE_RESOURCE) ? 1 : 2);
                            serviceBuilder.setSublibraryPriority(subPriority(library));
                            serviceBuilder.setMaterials(materials);
                            if (library.getDomainName() != null) {
                                serviceBuilder.setSource(library.getDomainName());
                            }
                            serviceBuilder.setSourceId(identifier);
                        } else {
                            logger.log(Level.WARNING, "ISIL not found in configuration: " + isil);
                        }
                        break;
                    case "c":
                        // skip domain, use the configured one derived from our library setup
                        break;
                    case "d":
                        // loan indicator
                        switch (subfield.getValue()) {
                            case "L", "a" -> serviceBuilder.setMode(Mode.LOAN);
                            case "C", "b" -> serviceBuilder.setMode(Mode.COPY);
                            case "LC", "c" -> serviceBuilder.setMode(Mode.LOAN_COPY);
                            case "N", "d" -> serviceBuilder.setMode(Mode.NONE);
                            case "e" -> serviceBuilder.setMode(Mode.ELECTRONIC_COPY);
                            default -> {
                            }
                        }
                        break;
                    case "e":
                        // second loan indicator
                        if ("p".equals(subfield.getValue())) {
                            serviceBuilder.setServicePriority(3);
                        }
                        break;
                    case "g":
                        location.setCallnumber(subfield.getValue());
                        break;
                    case "k":
                        // URL
                        if (!url.isEmpty()) {
                            location.addLink(url.toString(), label.toString());
                            url.setLength(0);
                            label.setLength(0);
                        }
                        url.append(subfield.getValue());
                        break;
                    case "l":
                        // URL label
                        label.append(subfield.getValue());
                        break;
                    case "z":
                        // textual holdings
                        location.setTextualHoldings(trimBrackets(subfield.getValue()));
                        break;
                    default:
                        break;
                }
            });
            if (!url.isEmpty()) {
                location.addLink(url.toString(), label.toString());
                url.setLength(0);
                label.setLength(0);
            }
            serviceBuilder.addLocation(location);
            list.add(serviceBuilder.build());
        });
        return list;
    }

    protected String identifier(MarcRecord marcRecord) {
        return trim(marcRecord.getFirst("001").getValue());
    }

    private static String trim(String string) {
        return string != null ? string.trim() : null;
    }

    private static String trimBrackets(String string) {
        return string.startsWith("<") && string.endsWith(">") ? string.substring(1, string.length() - 1) : string;
    }

    protected abstract void avail(AvailRequest request,
                                  AvailServiceResponseBuilder responseBuilder,
                                  MultiMap<String, LibraryService> multiMap,
                                  MarcRecord marcRecord);
}
