package org.xbib.interlibrary.z;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.action.resolve.AbstractResolveService;
import org.xbib.interlibrary.action.resolve.ResolveRequest;
import org.xbib.interlibrary.action.resolve.ResolveResponseBuilder;
import org.xbib.interlibrary.action.resolve.ResolveServiceResponse;
import org.xbib.interlibrary.action.resolve.ResolveServiceResponseBuilder;
import org.xbib.interlibrary.api.BibliographicDescription;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.interlibrary.common.DefaultBibliographicDescription;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.standardnumber.ZDB;
import org.xbib.z3950.client.api.Client;

public abstract class AbstractZResolveService extends AbstractResolveService {

    private static final Logger logger = Logger.getLogger(AbstractZResolveService.class.getName());

    protected final ZClientHelper clientHelper;

    protected final ReentrantLock lock;

    protected AbstractZResolveService(ServiceArguments arguments) {
        super(arguments);
        this.clientHelper = new ZClientHelper(arguments);
        this.lock = new ReentrantLock();
    }

    @Override
    public void execute(ResolveRequest request, ResolveResponseBuilder response) {
        ResolveServiceResponseBuilder serviceResponseBuilder = ResolveServiceResponse.builder();
        serviceResponseBuilder.setService(getName());
        try {
            lock.lock();
            serviceResponseBuilder.setRequest(request);
            BibliographicDescription bibliographicDescription = request.getBibliographicDescription();
            String source = bibliographicDescription.getSource();
            String sourceId = normalizeSourceId(bibliographicDescription.getSourceId());
            String formattedSourceId = bibliographicDescription.getFormattedSourceId() != null ?
                    bibliographicDescription.getFormattedSourceId() : bibliographicDescription.getSourceId();
            // we have to clean up the source ID to our reduced format in lower case and without hyphens
            // because it may come from external 'dirty' sources
            if ("zdb".equalsIgnoreCase(source)) {
                ZDB zdb = new ZDB();
                try {
                    boolean fixChecksum = bibliographicDescription.getChecksum();
                    if (fixChecksum) {
                        sourceId = sourceId + "0"; // add dummy checksum
                    }
                    zdb.set(sourceId).createChecksum(fixChecksum).normalize().verify();
                } catch (NumberFormatException e) {
                    logger.log(Level.WARNING, "invalid ZDB detected: " + sourceId);
                }
                sourceId = zdb.normalizedValue();
                formattedSourceId = zdb.format();
            }
            logger.log(Level.FINER, "original source id = " + bibliographicDescription.getSourceId() +
                    " normalized = " + sourceId +
                    " formatted = " + formattedSourceId +
                    " checksum = " + bibliographicDescription.getChecksum());
            DefaultBibliographicDescription.Builder bibliographicDescriptionBuilder = DefaultBibliographicDescription.builder()
                    .setSource(source)
                    .setSourceId(sourceId)
                    .setFormattedSourceId(formattedSourceId)
                    .setChecksum(bibliographicDescription.getChecksum())
                    .setIdentifiers(bibliographicDescription.getIdentifiers())
                    .setAuthor(bibliographicDescription.getAuthor())
                    .setTitle(bibliographicDescription.getTitle())
                    .setInTitle(bibliographicDescription.getInTitle())
                    .setPublisher(bibliographicDescription.getPublisher())
                    .setYear(bibliographicDescription.getYear())
                    .setVolume(bibliographicDescription.getVolume())
                    .setIssue(bibliographicDescription.getIssue())
                    .setPagination(bibliographicDescription.getPagination());
            String query = getResolveCQL(source, sourceId);
            if (query == null) {
                logger.log(Level.FINE, getName() + ": resolve query is null, rejected");
                return;
            }
            logger.log(Level.FINE, getName() + ": executing resolve request = " + request.asMap() + " CQL query = " + query);
            clientHelper.open();
            Client client = clientHelper.getClient();
            client.searchCQL(query, 0, 1, null,
                (status, total, returned, elapsedMillis) ->
                        logger.log(Level.FINE, "status = " + status + " total = " + total + " returned = " + returned + " millis = " + elapsedMillis),
                record -> {
                    String string = record.toString(StandardCharsets.UTF_8);
                    try (InputStream inputStream = new ByteArrayInputStream(string.getBytes(StandardCharsets.UTF_8))) {
                        String format = clientHelper.getFormat();
                        String type = clientHelper.getType();
                        Charset charset = Charset.forName(clientHelper.getEncoding());
                        Marc.Builder builder = Marc.builder()
                                .setFormat(format)
                                .setType(type)
                                .setInputStream(inputStream)
                                .setCharset(charset);
                        // XML?
                        if ("1.2.840.10003.5.109.10".equals(client.getPreferredRecordSyntax())) {
                            for (MarcRecord marcRecord : builder.xmlIterable()) {
                                resolve(request, serviceResponseBuilder, bibliographicDescriptionBuilder, marcRecord);
                            }
                        } else {
                            for (MarcRecord marcRecord : builder.iterable()) {
                                resolve(request, serviceResponseBuilder, bibliographicDescriptionBuilder, marcRecord);
                            }
                        }
                        serviceResponseBuilder.setBibliographicDescription(bibliographicDescriptionBuilder.build());
                        serviceResponseBuilder.setStatus(getName(), Response.Status.OK);
                    } catch (Exception e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                        throw new IllegalStateException(e);
                    }
                },
                () -> {
                  serviceResponseBuilder.setStatus(getName(), Response.Status.TIMEOUT);
                  logger.log(Level.WARNING, getName() + ": timeout");
                });
            clientHelper.close();
        } catch (Throwable e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            serviceResponseBuilder.setStatus(getName(), Response.Status.ERROR);
            throw new IllegalStateException(e);
        } finally {
            response.add(serviceResponseBuilder);
            lock.unlock();
        }
    }

    @Override
    public void close() throws IOException {
    }

    protected String normalizeSourceId(String sourceId) {
        return sourceId != null ? sourceId.toLowerCase(Locale.ROOT).trim().replace("-", "") : null;
    }

    protected String getResolveCQL(String source, String sourceId) {
        String supportedSource = supports(source);
        if (supportedSource == null) {
            return null;
        }
        return "bib.identifierLocalNumber = " + sourceId;
    }

    protected abstract void resolve(ResolveRequest resolveRequest,
                                    ResolveServiceResponseBuilder serviceResponseBuilder,
                                    DefaultBibliographicDescription.Builder builder,
                                    MarcRecord marcRecord);


    protected abstract String supports(String source);
}
