package org.xbib.interlibrary.z;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ConnectException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.text.Normalizer;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import org.xbib.interlibrary.action.search.AbstractSearchService;
import org.xbib.interlibrary.action.search.SearchRequest;
import org.xbib.interlibrary.action.search.SearchResponse;
import org.xbib.interlibrary.action.search.SearchResponseBuilder;
import org.xbib.interlibrary.action.search.SearchService;
import org.xbib.interlibrary.action.search.SearchSyntax;
import org.xbib.interlibrary.api.action.Response;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.marc.Marc;
import org.xbib.marc.MarcRecord;
import org.xbib.z3950.client.api.Client;
import org.xbib.z3950.common.operations.SortOperation;

import static org.xbib.interlibrary.action.search.SearchSyntax.CQL;
import static org.xbib.interlibrary.action.search.SearchSyntax.PQF;

public abstract class AbstractZSearchService extends AbstractSearchService<SearchRequest, SearchResponse<SearchRequest>, SearchService> {

    private static final Logger logger = Logger.getLogger(AbstractZSearchService.class.getName());

    private static final Pattern wordPattern = Pattern.compile("(\\p{L}\\p{M}*)+");

    private static final Pattern punctuationPattern = Pattern.compile("(?U)\\p{Punct}+");

    protected final MarcHelper marcHelper;

    private ZClientHelper zClientHelper;

    protected AbstractZSearchService(ServiceArguments arguments) {
        super(arguments);
        this.marcHelper = MarcHelper.INSTANCE;
    }

    @Override
    public void execute(SearchRequest request, SearchResponseBuilder response) {
        zClientHelper = new ZClientHelper(arguments);
        try {
            zClientHelper.open();
            final Client client = zClientHelper.getClient();
            SearchSyntax searchSyntax = evaluateSyntax(request);
            switch (searchSyntax) {
                case NATIVE, PQF -> {
                    // PQF format: <field + attributes> <querystring>
                    String pqf = encodePQF(request);
                    if (pqf.isEmpty()) {
                        logger.log(Level.WARNING, "unable to execute empty query");
                        response.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
                        return;
                    }
                    // sort?
                    List<SortOperation.SortParameter> sort = encodeSort(request);
                    // fix offset by one?
                    int offset = fixByOne() ? request.getOffset() + 1 : request.getOffset();
                    logger.log(Level.FINE, "pqf = " + pqf +
                            " fixbyone = " + fixByOne() +
                            " offset = " + offset + " size = " + request.getSize() +
                            " sort = " + request.getSortField());
                    client.searchPQF(pqf, offset, request.getSize(), sort,
                            (status, total, returned, elapsedMillis) -> {
                                response.addMetadata(getName(), "total", (long) total);
                                if (total > 0) {
                                    response.addMetadata(getName(), "took", elapsedMillis);
                                    response.addMetadata(getName(), "age", request.getAge());
                                    response.setStatus(getName(), Response.Status.OK);
                                } else {
                                    response.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
                                }
                            },
                            record -> {
                                String encoding = zClientHelper.getEncoding();
                                if (encoding == null) {
                                    encoding = "UTF-8";
                                }
                                String content = record.toString(Charset.forName(encoding));
                                // XML?
                                if ("1.2.840.10003.5.109.10".equals(client.getPreferredRecordSyntax())) {
                                    try (InputStream inputStream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8))) {
                                        for (MarcRecord marcRecord : Marc.builder()
                                                .setInputStream(inputStream)
                                                .setCharset(StandardCharsets.UTF_8)
                                                .xmlIterable()) {
                                            search(request, response, record.getNumber(), marcRecord);
                                        }
                                    } catch (Exception e) {
                                        logger.log(Level.SEVERE, e.getMessage(), e);
                                        response.setStatus(getName(), Response.Status.ERROR);
                                        response.setThrowable(getName(), e.getMessage(), e);
                                        throw new IllegalStateException(e);
                                    }
                                } else {
                                    try (InputStream inputStream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8))) {
                                        String format = zClientHelper.getFormat();
                                        String type = zClientHelper.getType();
                                        Charset charset = StandardCharsets.UTF_8;
                                        Marc.Builder marcBuilder = Marc.builder()
                                                .setFormat(format)
                                                .setType(type)
                                                .setInputStream(inputStream)
                                                .setCharset(charset);
                                        for (MarcRecord marcRecord : marcBuilder.iterable()) {
                                            search(request, response, record.getNumber(), marcRecord);
                                        }
                                    } catch (Exception e) {
                                        logger.log(Level.SEVERE, e.getMessage(), e);
                                        response.setStatus(getName(), Response.Status.ERROR);
                                        response.setThrowable(getName(), e.getMessage(), e);
                                        throw new IllegalStateException(e);
                                    }
                                }
                            },
                            () -> {
                                zClientHelper.timeout();
                                response.addMetadata(getName(), "total", 0L);
                                response.setStatus(getName(), Response.Status.TIMEOUT);
                            });
                }
                case CQL -> {
                    String cql = encodeCQL(request);
                    if (cql.isEmpty()) {
                        logger.log(Level.WARNING, "unable to execute empty query");
                        response.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
                        return;
                    }
                    // sort?
                    List<SortOperation.SortParameter> sort = encodeSort(request);
                    int offset = fixByOne() ? request.getOffset() + 1 : request.getOffset();
                    logger.log(Level.FINE, "cql = " + cql +
                            " fixbyone = " + fixByOne() +
                            " offset = " + offset + " size = " + request.getSize() +
                            " sort = " + request.getSortField() + " dir = " + request.getSortDirection());
                    client.searchCQL(cql, offset, request.getSize(), sort,
                            (status, total, returned, elapsedMillis) -> {
                                response.addMetadata(getName(), "total", (long) total);
                                if (total > 0) {
                                    response.addMetadata(getName(), "took", elapsedMillis);
                                    response.addMetadata(getName(), "age", request.getAge());
                                    response.setStatus(getName(), Response.Status.OK);
                                } else {
                                    response.setStatus(getName(), Response.Status.EMPTY_RESPONSE);
                                }
                            },
                            record -> {
                                String content = record.toString(StandardCharsets.UTF_8);
                                // XML?
                                if ("1.2.840.10003.5.109.10".equals(client.getPreferredRecordSyntax())) {
                                    try (InputStream inputStream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8))) {
                                        for (MarcRecord marcRecord : Marc.builder()
                                                .setInputStream(inputStream)
                                                .setCharset(StandardCharsets.UTF_8)
                                                .xmlIterable()) {
                                            search(request, response, record.getNumber(), marcRecord);
                                        }
                                    } catch (Exception e) {
                                        logger.log(Level.SEVERE, e.getMessage(), e);
                                        response.setStatus(getName(), Response.Status.ERROR);
                                        response.setThrowable(getName(), e.getMessage(), e);
                                        throw new IllegalStateException(e);
                                    }
                                } else {
                                    try (InputStream inputStream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8))) {
                                        String format = zClientHelper.getFormat();
                                        String type = zClientHelper.getType();
                                        Charset charset = Charset.forName(zClientHelper.getEncoding());
                                        Marc.Builder marcBuilder = Marc.builder()
                                                .setFormat(format)
                                                .setType(type)
                                                .setInputStream(inputStream)
                                                .setCharset(charset);
                                        for (MarcRecord marcRecord : marcBuilder.iterable()) {
                                            search(request, response, record.getNumber(), marcRecord);
                                        }
                                    } catch (Exception e) {
                                        logger.log(Level.SEVERE, e.getMessage(), e);
                                        response.setStatus(getName(), Response.Status.ERROR);
                                        response.setThrowable(getName(), e.getMessage(), e);
                                        throw new IllegalStateException(e);
                                    }
                                }
                            },
                            () -> {
                                zClientHelper.timeout();
                                response.addMetadata(getName(), "total", 0L);
                                response.setStatus(getName(), Response.Status.TIMEOUT);
                            });
                }
            }
        } catch (ConnectException e) {
            logger.log(Level.WARNING, e.getMessage());
            response.setStatus(getName(), Response.Status.ERROR);
            response.setThrowable(getName(), e.getMessage(), e);
        } catch (Exception e) {
            if (e.getMessage().contains("Read timed out") || e.getMessage().contains("Connect timed out")) {
                response.setStatus(getName(), Response.Status.TIMEOUT);
                zClientHelper.timeout();
            } else {
                logger.log(Level.SEVERE, e.getMessage(), e);
                response.setStatus(getName(), Response.Status.ERROR);
                response.setThrowable(getName(), e.getMessage(), e);
            }
        } finally {
            zClientHelper.close();
        }
    }

    @Override
    public void close() throws IOException {
        super.close();
    }

    @Override
    public boolean supports(SearchSyntax searchSyntax) {
        return PQF == searchSyntax || CQL == searchSyntax;
    }

    protected SearchSyntax evaluateSyntax(SearchRequest searchRequest) {
        if (searchRequest.getSearchKey().contains("cql")) {
            return CQL;
        } else {
            return searchRequest.getSearchSyntax();
        }
    }

    protected String encodeCQL(SearchRequest request) {
        // abbreviated: single query parameter
        if (request.getSearchKey().size() == 1 &&
                request.getSearchKey().get(0).isEmpty() &&
                request.getSearchValue().size() == 1) {
            return trim(request.getSearchValue().get(0));
        }
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < request.getSearchValue().size(); i++) {
            String key = request.getSearchKey().get(i);
            String value = request.getSearchValue().get(i);
            String op = request.getSearchOp().get(i);
            if (!sb.isEmpty() && !op.isEmpty() && !value.isEmpty()) {
                sb.append(' ').append(op).append(' ');
            }
            sb.append(buildCQLQueryTerm(key, value));
        }
        return sb.toString();
    }

    protected String buildCQLQueryTerm(String key, String query) {
        if (query == null || query.isEmpty()) {
            return "";
        }
        String encodedQuery = encodeQueryString(query);
        if (key != null && !key.isEmpty() && !encodedQuery.isEmpty()) {
            encodedQuery = key + " " + getRelation() + " " + encodedQuery;
        }
        return encodedQuery;
    }

    protected String getRelation() {
        return "all";
    }

    protected String encodePQF(SearchRequest request) {
        // abbreviated: single query parameter
        if (request.getSearchKey().size() == 1 &&
                request.getSearchKey().get(0).isEmpty() &&
                request.getSearchValue().size() == 1) {
            return trim(request.getSearchValue().get(0));
        }
        LinkedList<String> termList = new LinkedList<>();
        for (int i = 0; i < request.getSearchValue().size(); i++) {
            String value = request.getSearchValue().get(i);
            value = trim(value);
            if (!value.isEmpty()) {
                String key = request.getSearchKey().get(i);
                String term = "pqf".equals(key) ? value : buildPQFQueryTerm(key, value);
                if (term.isEmpty()) {
                    continue;
                }
                if (i > 0) {
                    String op = request.getSearchOp().get(i);
                    if (!op.isEmpty()) {
                        // get previous term if present
                        if (!termList.isEmpty()) {
                            String previous = termList.pop();
                            termList.push(op + " " + previous + " " + term);
                        } else {
                            termList.push(term);
                        }
                    }
                } else {
                    termList.push(term);
                }
            }
        }
        return String.join(" ", termList);
    }

    protected String buildPQFQueryTerm(String key, String query) {
        String attrs = zClientHelper.getAttributes().get(key);
        String encodedQuery = encodeQueryString(query);
        if (attrs != null && !attrs.isEmpty() && !encodedQuery.isEmpty()) {
            encodedQuery = attrs + " " + encodedQuery;
        }
        return encodedQuery;
    }

    protected List<SortOperation.SortParameter> encodeSort(SearchRequest searchRequest) {
        if (searchRequest.getSortField() != null && !searchRequest.getSortField().isEmpty()) {
            String sortField = zClientHelper.getSort().get(searchRequest.getSortField() + ".field");
            boolean sortDirection = zClientHelper.getSort().getAsBoolean(searchRequest.getSortField() + ".ascending", true);
            if (sortField != null) {
                return List.of(SortOperation.SortParameter.of(sortField)
                        .setAscending(sortDirection));
            }
        }
        return null;
    }

    protected String encodeQueryStringToAscii(String key, String value) {
        String result = value;
        result = result
                .replaceAll("ä", "ae")
                .replaceAll("ö", "oe")
                .replaceAll("ü", "ue")
                .replaceAll("Ä", "Ae")
                .replaceAll("Ö", "Oe")
                .replaceAll("Ü", "Ue")
                .replaceAll("ß", "ss");
        if ("author".equals(key)) {
            // force blank after comma
            result = result.replaceAll("(\\S+),(\\S+)", "\\1, \\2");
        }
        if (key.contains("title")) {
            result = result.replaceAll("(\\S+)-(\\S+)", "\\1#\\2");
        }
        result = Normalizer.normalize(result, Normalizer.Form.NFKD);
        result = result.replaceAll("[^\\x00-\\x7F]", "");
        if ("completetitle".equals(key)) {
            result = "'" + result + "'";
        }
        return result.startsWith("\"") && result.endsWith("\"") ? result : "\"" + result + "\"";
    }

    protected String encodeQueryString(String value) {
        String normalizedValue = normalizeValue(value);
        return isSingleWord(value) ? normalizedValue : quote(normalizedValue);
    }

    protected String normalizeValue(String value) {
        if (matchesPunctuation(value)) {
            return value.replaceAll("(?U)\\p{Punct}+", "");
        }
        return value;
    }

    protected static String trim(String string) {
        return string != null ? string.trim() : "";
    }

    protected static boolean isSingleWord(String string) {
        return string != null && wordPattern.matcher(string).matches();
    }

    protected static boolean matchesPunctuation(String string) {
        return string != null && punctuationPattern.matcher(string).matches();
    }

    protected static String quote(String string) {
        return string.isEmpty() ? string : '"' + getEscapedString(string) + '"';
    }

    protected static String getEscapedString(String value) {
		if (value == null) {
			return null;
		}
		return value
				.replace("\\", "\\\\")
				.replace("\n", "\\n")
				.replace("\t", "\\t")
				.replace("\b", "\\b")
				.replace("\r", "\\r")
				.replace("\f", "\\f")
				.replace("\"", "\\\"")
				.replace("'", "\\'");
	}

    protected abstract void search(SearchRequest request,
                                   SearchResponseBuilder response,
                                   int recordNumber,
                                   MarcRecord marcRecord);
}
