package org.xbib.interlibrary.z;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.datastructures.tiny.TinyList;
import org.xbib.datastructures.tiny.TinyMap;
import org.xbib.interlibrary.api.library.Material;
import org.xbib.interlibrary.common.util.InputStreamHelper;
import org.xbib.marc.MarcField;
import org.xbib.marc.MarcRecord;
import org.xbib.marc.label.MultipartResourceRecordLevel;
import org.xbib.marc.label.RecordLabel;

import static org.xbib.interlibrary.api.library.Material.ELECTRONIC_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.GLOBE;
import static org.xbib.interlibrary.api.library.Material.KIT;
import static org.xbib.interlibrary.api.library.Material.MAP;
import static org.xbib.interlibrary.api.library.Material.MICROFORM;
import static org.xbib.interlibrary.api.library.Material.MOTION_PICTURE;
import static org.xbib.interlibrary.api.library.Material.NON_PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.NOTATED_MUSIC;
import static org.xbib.interlibrary.api.library.Material.ONLINE_RESOURCE;
import static org.xbib.interlibrary.api.library.Material.PROJECTED_GRAPHIC;
import static org.xbib.interlibrary.api.library.Material.REMOTE_SENSING_IMAGE;
import static org.xbib.interlibrary.api.library.Material.SOUND_RECORDING;
import static org.xbib.interlibrary.api.library.Material.TACTILE;
import static org.xbib.interlibrary.api.library.Material.TEXT;
import static org.xbib.interlibrary.api.library.Material.UNSPECIFIED;
import static org.xbib.interlibrary.api.library.Material.VIDEO_RECORDING;

public final class MarcHelper {

    private static final Logger logger = Logger.getLogger(MarcHelper.class.getName());

    public static final MarcHelper INSTANCE = new MarcHelper();

    private final Map<String, Object> country;

    private final Map<String, Object> language;

    private MarcHelper() {
        this.country = TinyMap.builder();
        country.putAll(Objects.requireNonNull(getMap("/org/xbib/interlibrary/z/marc/country.json")));
        this.language = TinyMap.builder();
        language.putAll(Objects.requireNonNull(getMap("/org/xbib/interlibrary/z/marc/language.json")));
    }

    private static Map<String, Object> getMap(String path) {
        try {
            return Json.toMap(InputStreamHelper.readString(AbstractZSearchService.class.getResourceAsStream(path), StandardCharsets.UTF_8));
        } catch (IOException e) {
            logger.log(Level.WARNING, e.getMessage(), e);
            return null;
        }
    }

    public Supplier<String> getRecordIdSupplier(MarcRecord marcRecord) {
        final String id = extractIdentifier(marcRecord, "001");
        return () -> id;
    }

    public Supplier<String> getPrefixIdSupplier(MarcRecord marcRecord) {
        String id = extractIdentifier(marcRecord, "001");
        String prefix = extractIdentifier(marcRecord, "003");
        final String prefixedId = prefix != null && !Objects.requireNonNull(id).startsWith("(" + prefix + ")") ? "(" + prefix + ")" + id : id;
        return () -> prefixedId;
    }

    public Supplier<String> getZdbIdOrPrefixSupplier(MarcRecord marcRecord) {
        final String id = extractZdbIdentifier(marcRecord, "035");
        if (id != null) {
            return () -> id;
        }
        return getPrefixIdSupplier(marcRecord);
    }

    public String extractIdentifier(MarcRecord marcRecord, String identifierField) {
        String string = marcRecord.getAll(identifierField).stream().findFirst()
                .map(MarcField::getValue).orElse(null);
        return string == null ? null : string.trim();
    }

    public Map<String, String> extractIdentifiers(MarcRecord marcRecord, String identifierFeld) {
        TinyMap.Builder<String, String> builder = TinyMap.builder();
        marcRecord.getAll(identifierFeld)
                .forEach(mf -> builder.put(prefixOf(mf.getFirstSubfieldValue("a")), suffixOf(mf.getFirstSubfieldValue("a"))));
        return builder.build();
    }

    public String extractZdbIdentifier(MarcRecord marcRecord, String identifierField) {
        StringBuilder sb = new StringBuilder();
        marcRecord.getAll(identifierField)
                .stream().filter(mf -> mf.getFirstSubfieldValue("a").startsWith("(DE-599)ZDB"))
                .forEach(mf -> sb.append(mf.getFirstSubfieldValue("a").substring(11)));
        return sb.isEmpty() ? null : sb.toString();
    }

    public String extractDate(MarcRecord marcRecord) {
        return marcRecord.getAll("008").stream().findFirst()
                .map(MarcField::getValue).orElse("       0000").substring(7, 11);
    }

    public String extractChronology(MarcRecord marcRecord, String field, String subfieldId) {
        return marcRecord.getAll(field).stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue(subfieldId)).orElse(null);
    }

    public List<String> extractCreators(MarcRecord marcRecord, List<String> creatorFields) {
        List<String> list = new ArrayList<>();
        creatorFields.forEach(creatorFieldName ->
                marcRecord.getAll(creatorFieldName).forEach(f ->
                        list.addAll(f.getSubfields().stream()
                                .filter(sf -> filterCreator(sf.getValue()))
                                .map(sf -> cleanCreator(sf.getValue()))
                                .filter(MarcHelper::notEmpty)
                                .toList())));
        return list;
    }

    public List<String> extractTitle(MarcRecord marcRecord, List<String> titleFields) {
        List<String> list = new ArrayList<>();
        titleFields.forEach(titleField ->
                marcRecord.getAll(titleField).forEach(f ->
                        list.addAll(f.getSubfields().stream()
                                .filter(sf -> !"6".equals(sf.getId())) // remove "$6 880-01" transliteration content
                                .map(sf -> cleanTitle(sf.getValue()))
                                .filter(MarcHelper::notEmpty)
                                .toList())));
        return list;
    }

    public String extractCountry(MarcRecord marcRecord, String countryField, String subfieldId) {
        return marcRecord.getAll(countryField).stream().findFirst()
                .map(mf -> (String) country.get(mf.getFirstSubfieldValue(subfieldId))).orElse(null);
    }

    public String extractLanguage(MarcRecord marcRecord, String languageField, String subfieldId) {
        return marcRecord.getAll(languageField).stream().findFirst()
                .map(mf -> (String) language.get(mf.getFirstSubfieldValue(subfieldId))).orElse(null);
    }

    public List<String> extractPublication(MarcRecord marcRecord, Collection<String> fields, String subfieldId) {
        List<String> list = new ArrayList<>();
        fields.forEach(field ->
                marcRecord.getAll(field).stream()
                        .map(mf -> mf.getFirstSubfieldValue(subfieldId))
                        .filter(Objects::nonNull)
                        .filter(s -> !s.isEmpty())
                        .forEach(list::add));
        return list;
    }

    public List<String> extractPublishedAt(MarcRecord marcRecord, Collection<String> fields, String subfieldId) {
        List<String> list = new ArrayList<>();
        fields.forEach(field ->
            marcRecord.getAll(field).stream()
                    .map(mf -> mf.getFirstSubfieldValue(subfieldId))
                    .filter(Objects::nonNull)
                    .filter(s -> !s.isEmpty())
                    .forEach(list::add));
        return list;
    }

    public List<String> extractPublishedBy(MarcRecord marcRecord, Collection<String> fields, String subfieldId) {
        List<String> list = new ArrayList<>();
        fields.forEach(field ->
            marcRecord.getAll(field).stream()
                    .map(mf -> mf.getFirstSubfieldValue(subfieldId))
                    .filter(Objects::nonNull)
                    .filter(s -> !s.isEmpty())
                    .forEach(list::add));
        return list;
    }

    public List<String> extractPublishedOn(MarcRecord marcRecord, Collection<String> fields, String subfieldId) {
        List<String> list = new ArrayList<>();
        fields.forEach(field ->
            marcRecord.getAll(field).stream()
                    .map(mf -> mf.getFirstSubfieldValue(subfieldId))
                    .filter(Objects::nonNull)
                    .filter(s -> !s.isEmpty())
                    .forEach(list::add));
        return list;
    }

    public Collection<String> extractIBN(MarcRecord marcRecord, String field, List<String> indicators, List<String> subfieldIds) {
        Set<String> set = new LinkedHashSet<>();
        marcRecord.getAll(field).stream()
                .filter(mf -> indicators == null || indicators.contains(mf.getIndicator())).forEach(f ->
                        f.getSubfields().stream()
                                .filter(sf -> subfieldIds == null || subfieldIds.contains(sf.getId()))
                                .map(MarcField.Subfield::getValue)
                                .filter(Objects::nonNull)
                                .filter(s -> !s.isEmpty())
                                .forEach(set::add));
        return set;
    }

    public String extractFirstDate(MarcRecord marcRecord, String field, String subfieldId) {
        return marcRecord.getAll(field).stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue(subfieldId)).orElse(null);
    }

    public String extractExtent(MarcRecord marcRecord, String field, String subfieldId) {
        return marcRecord.getAll(field).stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue(subfieldId)).orElse(null);
    }

    public List<String> extractMaterialAsString(MarcRecord marcRecord) {
        return marcRecord.getAll("007").stream()
                .map(MarcField::recoverControlFieldValue)
                .map(this::toMaterial)
                .map(Material::toValue)
                .toList();
    }

    public List<Material> extractMaterial(MarcRecord marcRecord) {
        return marcRecord.getAll("007").stream()
                .map(MarcField::recoverControlFieldValue)
                .map(this::toMaterial)
                .toList();
    }

    private Material toMaterial(String value) {
        return switch (value.substring(0, 1)) {
            case "a" -> MAP;
            case "c" -> value.startsWith("cr") ? ONLINE_RESOURCE :ELECTRONIC_RESOURCE;
            case "d" -> GLOBE;
            case "f" -> TACTILE;
            case "g" -> PROJECTED_GRAPHIC;
            case "h" -> MICROFORM;
            case "k" -> NON_PROJECTED_GRAPHIC;
            case "m" -> MOTION_PICTURE;
            case "o" -> KIT;
            case "q" -> NOTATED_MUSIC;
            case "r" -> REMOTE_SENSING_IMAGE;
            case "s" -> SOUND_RECORDING;
            case "t" -> TEXT;
            case "v" -> VIDEO_RECORDING;
            default -> UNSPECIFIED;
        };
    }

    public boolean extractMultipart(MarcRecord marcRecord) {
        return marcRecord.getRecordLabel().getMultipartResourceRecordLevel() == MultipartResourceRecordLevel.SET;
    }

    public List<String> extractFormatCodes(MarcRecord marcRecord) {
        List<String> list = new ArrayList<>();
        marcRecord.getAll("336").stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue("a"))
                .ifPresent(list::add);
        marcRecord.getAll("337").stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue("a"))
                .ifPresent(list::add);
        marcRecord.getAll("338").stream().findFirst()
                .map(mf -> mf.getFirstSubfieldValue("a"))
                .ifPresent(list::add);
        return list;
    }

    public List<String> extractTypeCodes(MarcRecord marcRecord) {
        List<String> list = new ArrayList<>();
        RecordLabel recordLabel = marcRecord.getRecordLabel();
        if (recordLabel == null) {
            throw new IllegalArgumentException("no record label?");
        }
        switch (recordLabel.getTypeOfRecord()) {
            case LANGUAGE_MATERIAL -> list.add("Sprachmaterialien");
            case LANGUAGE_MATERIAL_MANUSCRIPT, LANGUAGE_MATERIAL_MANUSCRIPT_MARC21 ->
                    list.add("Handschriftliche Sprachmaterialien");
            case NOTATED_MUSIC -> list.add("Noten");
            case NOTATED_MUSIC_MANUSCRIPT -> list.add("Handschriftliche Noten");
            case CARTOGRAPHIC_MATERIAL -> list.add("Kartografische Materialien");
            case CARTOGRAPHIC_MATERIAL_MANUSCRIPT -> list.add("Handschriftliche, kartografische Materialien");
            case PROJECTED_MEDIUM -> list.add("Projiziertes Medium");
            case NONMUSICAL_SOUND_RECORDING -> list.add("Nicht-musikalische Tonaufnahme");
            case MUSICAL_SOUND_RECORDING -> list.add("Musikaufnahme");
            case PICTURE -> list.add("Zweidimensionale nicht-projizierte Grafik");
            case ELECTRONIC_RESOURCE -> list.add("Elektronische Ressource");
            case COMPUTER_FILE -> list.add("Datei");
            case KIT -> list.add("Medienkombination");
            case MIXED_MATERIALS -> list.add("Verschiedene Materialien");
            case ARTIFACT -> list.add("Dreidimensionales Artefakt oder natürlicher Gegenstand");
            default -> {
            }
        }
        return list;
    }

    public List<Map<String, Object>> extractLinks(MarcRecord marcRecord) {
        TinyList.Builder<Map<String, Object>> list = TinyList.builder();
        marcRecord.getAll("856").forEach(marcField -> {
            TinyMap.Builder<String, Object> map = TinyMap.builder();
            String uri = marcField.getFirstSubfieldValue("u");
            if (uri == null) {
                uri = marcField.getFirstSubfieldValue("d");
            }
            if (uri == null) {
                uri = marcField.getFirstSubfieldValue("f");
            }
            putIf(map, "uri", uri);
            putIf(map, "publicnote", marcField.getFirstSubfieldValue("x"));
            putIf(map, "nonpublicnote", marcField.getFirstSubfieldValue("3"));
            list.add(map.build());
        });
        return list.build();
    }

    public List<Map<String, Object>> extractEmbedments(MarcRecord marcRecord) {
        TinyList.Builder<Map<String, Object>> list = TinyList.builder();
        marcRecord.getAll("773").stream()
                .filter(mf -> mf.getIndicator().equals("08")).forEach(marcField -> {
                    TinyMap.Builder<String, Object> map = TinyMap.builder();
                    putIf(map, "prefix", marcField.getFirstSubfieldValue("i"));
                    putIf(map, "title", marcField.getFirstSubfieldValue("t"));
                    putIf(map, "publisher", marcField.getFirstSubfieldValue("d"));
                    putIf(map, "pagination", marcField.getFirstSubfieldValue("g"));
                    putIf(map, "identifier", marcField.getFirstSubfieldValue("w"));
                    putIf(map, "issn", marcField.getFirstSubfieldValue("x"));
                    list.add(map.build());
                });
        return list.build();
    }

    public boolean filterCreator(String string) {
        return true;
    }

    public String cleanCreator(String string) {
        return string == null ? null : string.trim()
                .replaceAll("<<(.*?)>>", "\u0098$1\u009C")
                .replaceAll("<(.*?)>", "[$1]")
                .replaceAll("¬(.*?)¬", "\u0098$1\u009C")
                .replaceAll("VerfasserIn", "")
                .replaceAll("aut", "");
    }

    public static String cleanTitle(String string) {
        // u00ac = ¬
        return string == null ? null : string.trim()
                .replaceAll("<<(.*?)>>", "\u0098$1\u009C")
                .replaceAll("<(.*?)>", "[$1]")
                .replaceAll("¬(.*?)¬", "\u0098$1\u009C")
                .replaceAll("Elektronische Ressource", "");
    }

    public static void putIf(Map<String, Object> map, String key, String value) {
        if (key != null && value != null && !"null".equals(value)) {
            map.put(key, value);
        }
    }

    public static boolean notEmpty(String string) {
        return string != null && !string.isEmpty();
    }

    private static String prefixOf(String string) {
        if (string == null) {
            return "";
        }
        int pos1 = string.indexOf('(');
        int pos2 = string.indexOf(')');
        return pos1 >=0 && pos2 >= 0 ? string.substring(pos1 + 1, pos2) : "";
    }

    private static String suffixOf(String string) {
        if (string == null) {
            return null;
        }
        int pos1 = string.indexOf('(');
        int pos2 = string.indexOf(')');
        return pos1 >=0 && pos2 >= 0 ? string.substring(pos2 + 1) : string;
    }
}
