package org.xbib.interlibrary.z;

import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.interlibrary.api.action.ServiceArguments;
import org.xbib.settings.Settings;
import org.xbib.z3950.client.api.Client;
import org.xbib.z3950.client.jdk.JDKZClient;

public class ZClientHelper {

    private static final Logger logger = Logger.getLogger(ZClientHelper.class.getName());

    private final Settings settings;

    private final Settings attrs;

    private final Settings sort;

    private Client client;

    private final AtomicBoolean timeout;

    private final AtomicBoolean closed;

    public ZClientHelper(ServiceArguments arguments) {
        this((arguments.getSettings().getAsSettings("z3950." + arguments.getName())));
    }

    public ZClientHelper(Settings settings) {
        this.settings = settings.getAsSettings("connection");
        this.attrs = settings.getAsSettings("attr");
        this.sort = settings.getAsSettings("sort");
        this.closed = new AtomicBoolean(true);
        this.timeout = new AtomicBoolean(false);
    }

    public void close() {
        if (timeout.get()) {
            logger.log(Level.WARNING, "timeout, not closing Z client " + client + "  host " + client.getHost() + ":" + client.getPort());
            return;
        }
        if (closed.compareAndSet(false, true)) {
            logger.log(Level.FINE, "closing Z client " + client + " host " + client.getHost() + ":" + client.getPort());
            try {
                client.close();
            } catch (IOException e) {
                throw new IllegalStateException("unable to close Z client " + client + " host " + client.getHost() + ":" + client.getPort());
            }
            client = null; // for GC
        }
    }

    public String getFormat() {
        return settings.get("format");
    }

    public String getType() {
        return settings.get("type");
    }

    public String getEncoding() {
        return settings.get("encoding");
    }

    public Settings getAttributes() {
        return attrs;
    }

    public Settings getSort() {
        return sort;
    }

    public void open() {
        if (closed.compareAndSet(true, false)) {
            JDKZClient.Builder builder = JDKZClient.builder();
            builder.setImplementationName("Java Z Client")
                    .setImplementationVersion("1.00")
                    .setUser(settings.get("user"))
                    .setPass(settings.get("pass"))
                    .setHost(settings.get("host"))
                    .setPort(settings.getAsInt("port", -1))
                    .setDatabases(Arrays.asList(settings.getAsArray("database")))
                    .setResultSetName(settings.get("resultsetname"))
                    .setPreferredRecordSyntax(settings.get("preferredrecordsyntax", "marc21"))
                    .setEncoding(settings.get("encoding", "UTF-8"))
                    .setFormat(settings.get("format", "Marc21"))
                    .setType(settings.get("type", "Bibliographic"))
                    .setTimeout(settings.getAsLong("timeout", 5000L))
                    .wordListSupported(settings.getAsBoolean("wordListSupported", true));
            String elementSetName = settings.get("elementsetname");
            if (elementSetName != null) {
                builder.setElementSetName(elementSetName);
            }
            client = builder.build();
            logger.log(Level.FINE, "created Z client " + client +
                    " to host " + client.getHost() + ":" + client.getPort() +
                    " with timeout millis = " + client.getTimeout() +
                    " word list supported = " + settings.getAsBoolean("wordListSupported", true));
        }
    }

    public void timeout() {
        if (timeout.compareAndSet(false, true)) {
            logger.log(Level.FINE, "timeout of Z client " + client + " to host " + client.getHost() + ":" + client.getPort());
        }
    }

    public Client getClient() {
        return client;
    }
}
